#include <wchar.h>
#include <windows.h>
#include <stdio.h>
#include <cmath>
#include "NET_worldsplus_MumbleLink.h"

struct LinkedMem {
    UINT32	uiVersion;
    DWORD	uiTick;
    UINT32	context_len;
    float	fAvatarPosition[3];
    float	fAvatarFront[3];
    float	fAvatarTop[3];
    wchar_t	name[256];
    float	fCameraPosition[3];
    float	fCameraFront[3];
    float	fCameraTop[3];
    wchar_t	identity[256];
    unsigned char context[256];
    wchar_t description[2048];
};

LinkedMem *lm = NULL;
JNIEXPORT void JNICALL Java_NET_worldsplus_MumbleLink_initMumble(JNIEnv *env, jobject object) {
    HANDLE hMapObject = OpenFileMappingW(FILE_MAP_ALL_ACCESS, FALSE, L"MumbleLink");
    if (hMapObject == NULL)
        return;

    lm = (LinkedMem *) MapViewOfFile(hMapObject, FILE_MAP_ALL_ACCESS, 0, 0, sizeof(LinkedMem));
    if (lm == NULL) {
        CloseHandle(hMapObject);
        hMapObject = NULL;
        return;
    }
}

JNIEXPORT void JNICALL Java_NET_worldsplus_MumbleLink_updateMumble(JNIEnv *env, jobject object, jstring name, float posX, float posY, float posZ, float camX, float camY, float camZ, float yaw, float pitch, jstring room) {
    if (! lm)
        return;

    if(lm->uiVersion != 2) {
        wcsncpy(lm->name, L"WorldsMumbleLink", 256);
        wcsncpy(lm->description, L"WorldsMumbleLink is a link for Worlds.com", 2048);
        lm->uiVersion = 2;
    }
    lm->uiTick++;
    // Left handed coordinate system.
    // X positive towards "right".
    // Y positive towards "up".
    // Z positive towards "front".
    //
    // 1 unit = 1 meter

    // Unit vector pointing out of the avatar's eyes aka "At"-vector.
    lm->fAvatarFront[0] = cos(yaw)*cos(pitch);
    lm->fAvatarFront[1] = 0.0f;
    lm->fAvatarFront[2] = sin(yaw)*cos(pitch);

    // Unit vector pointing out of the top of the avatar's head aka "Up"-vector (here Top points straight up).
    lm->fAvatarTop[0] = 0.0f;
    lm->fAvatarTop[1] = 1.0f;
    lm->fAvatarTop[2] = 0.0f;

    // Position of the avatar
    lm->fAvatarPosition[0] = posX;
    lm->fAvatarPosition[1] = posZ;
    lm->fAvatarPosition[2] = posY;

    // Same as avatar but for the camera.
    lm->fCameraPosition[0] = camX;
    lm->fCameraPosition[1] = camZ;
    lm->fCameraPosition[2] = camY;

    lm->fCameraFront[0] = 0.0f;
    lm->fCameraFront[1] = 0.0f;
    lm->fCameraFront[2] = 1.0f;
    lm->fCameraTop[0] = 0.0f;
    lm->fCameraTop[1] = 1.0f;
    lm->fCameraTop[2] = 0.0f;

    // Identifier which uniquely identifies a certain player in a context (e.g. the ingame name).
    wcsncpy(lm->identity, ( wchar_t *)env->GetStringChars(name, NULL), 256);
    // Context should be equal for players which should be able to hear each other positional and
    // differ for those who shouldn't (e.g. it could contain the server+port and team)
    memcpy(lm->context, ( wchar_t *)env->GetStringChars(room, NULL), 16);
    lm->context_len = 16;
}
